#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Package definitions for the resource files.

see copyright/license https://github.com/DerwenAI/strwythura/README.md
"""

import pathlib

STRW_LOGO: pathlib.Path = pathlib.Path(__file__).resolve().parent / "logo.png"
SZ_LOGO: pathlib.Path = pathlib.Path(__file__).resolve().parent / "senzing.png"
